from ..exception import IncongruentFeature


class FeatureSet(object):

    def __init__(self, raw_dict):
        pass


def unify_features(l, r, match_required=None):
    """
    generate the union of left and right feature dictionaries., and
    raise an exception if there is a conflict in values.

    left, right are assumed to be dict<str, set()> objects

    if match_required is None, all keys must match.
    if match_required is not None, all keys found in the variable
    will require matching (empty collections are valid)
    """
    l = l.copy()
    r = r.copy()
    union = {}
    for key in set(l.keys()) | set(r.keys()):
        require_match = match_required is None or key in match_required
        union[key] = _combine_values(l.get(key), r.get(key),
                                     require_match=require_match)
    return union


def _combine_values(l_values, r_values, require_match=False):
    if not l_values:
        return r_values
    if not r_values:
        return l_values
    if require_match and l_values != r_values:
        raise IncongruentFeature(
            "{0} != {1}, and is a required match.".format(l_values, r_values)
        )
    return l_values | r_values



from nlgen.cfg import CFG, Terminal, ProductionRef, Production


def test_simple_production():
    cfg = CFG([
        ("S", Production([Terminal("foo"), Terminal("bar")]))
    ])
    expect = [("foo", "bar")]
    result = list(cfg.permutation_values("S"))
    assert expect == result



from nlgen.cfg.markup import parse_production
from nlgen.cfg import (
    Production, ProductionRef
)


def test_simple_case():
    result = parse_production("PRONOUN VI NOUN")
    assert result == Production([
        ProductionRef("PRONOUN"),
        ProductionRef("VI"),
        ProductionRef("NOUN"),
    ])



from nlgen.cfg import CFG, Terminal, ProductionRef


def test_simple_production_ref():
    cfg = CFG([
        ("S", Terminal("foo")),
    ])
    expect = [("foo",)]
    result = list(cfg.permutation_values(ProductionRef("S")))
    assert expect == result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# CAVEAT UTILITOR
#
# This file was automatically generated by Grako.
#
#    https://pypi.python.org/pypi/grako/
#
# Any changes you make to it will be overwritten the next time
# the file is generated.


from __future__ import print_function, division, absolute_import, unicode_literals

from grako.buffering import Buffer
from grako.parsing import graken, Parser
from grako.util import re, RE_FLAGS, generic_main  # noqa


__version__ = (2016, 9, 19, 4, 9, 53, 0)

__all__ = [
    'CFGMarkupParser',
    'CFGMarkupSemantics',
    'main'
]

KEYWORDS = set([])


class CFGMarkupBuffer(Buffer):
    def __init__(self,
                 text,
                 whitespace=None,
                 nameguard=None,
                 comments_re=None,
                 eol_comments_re=None,
                 ignorecase=None,
                 namechars='',
                 **kwargs):
        super(CFGMarkupBuffer, self).__init__(
            text,
            whitespace=whitespace,
            nameguard=nameguard,
            comments_re=comments_re,
            eol_comments_re=eol_comments_re,
            ignorecase=ignorecase,
            namechars=namechars,
            **kwargs
        )


class CFGMarkupParser(Parser):
    def __init__(self,
                 whitespace=None,
                 nameguard=None,
                 comments_re=None,
                 eol_comments_re=None,
                 ignorecase=None,
                 left_recursion=True,
                 parseinfo=True,
                 keywords=KEYWORDS,
                 namechars='',
                 **kwargs):
        super(CFGMarkupParser, self).__init__(
            whitespace=whitespace,
            nameguard=nameguard,
            comments_re=comments_re,
            eol_comments_re=eol_comments_re,
            ignorecase=ignorecase,
            left_recursion=left_recursion,
            parseinfo=parseinfo,
            keywords=keywords,
            namechars=namechars,
            **kwargs
        )

    def parse(self, text, *args, **kwargs):
        if not isinstance(text, Buffer):
            text = CFGMarkupBuffer(text, **kwargs)
        return super(CFGMarkupParser, self).parse(text, *args, **kwargs)

    @graken()
    def _reference_(self):
        self._pattern(r'[a-zA-Z0-9]+')
        self.name_last_node('key')
        self.ast._define(
            ['key'],
            []
        )

    @graken()
    def _terminal_(self):
        self._pattern(r'"[^"]*"')
        self.name_last_node('value')
        self.ast._define(
            ['value'],
            []
        )

    @graken()
    def _single_production_(self):
        with self._choice():
            with self._option():
                self._reference_()
            with self._option():
                self._terminal_()
            self._error('no available options')

    @graken()
    def _production_(self):

        def block1():
            self._single_production_()
        self._positive_closure(block1)
        self.name_last_node('production_list')
        self.ast._define(
            ['production_list'],
            []
        )


class CFGMarkupSemantics(object):
    def reference(self, ast):
        return ast

    def terminal(self, ast):
        return ast

    def single_production(self, ast):
        return ast

    def production(self, ast):
        return ast


def main(
        filename,
        startrule,
        trace=False,
        whitespace=None,
        nameguard=None,
        comments_re=None,
        eol_comments_re=None,
        ignorecase=None,
        left_recursion=True,
        parseinfo=True,
        **kwargs):

    with open(filename) as f:
        text = f.read()
    whitespace = whitespace or None
    parser = CFGMarkupParser(parseinfo=False)
    ast = parser.parse(
        text,
        startrule,
        filename=filename,
        trace=trace,
        whitespace=whitespace,
        nameguard=nameguard,
        ignorecase=ignorecase,
        **kwargs)
    return ast

if __name__ == '__main__':
    import json
    ast = generic_main(main, CFGMarkupParser, name='CFGMarkup')
    print('AST:')
    print(ast)
    print()
    print('JSON:')
    print(json.dumps(ast, indent=2))
    print()



import pytest
from nlgen.exception import IncongruentFeature
from nlgen.cfg.feature import unify_features


def test_feature():
    left = {"gender": set(["male"]), "person": set([1])}
    right = {"gender": set(["male"]), "person": set([1])}
    expect = {"gender": set(["male"]), "person": set([1])}
    assert unify_features(left, right) == expect


def test_empty_features():
    """ empty feature values on one side should default to the other side's values. """
    left = {"person": set([1])}
    right = {"gender": set(["male"])}
    expect = {"gender": set(["male"]), "person": set([1])}
    assert unify_features(left, right) == expect


def test_incongruent_feature():
    left = {"gender": set(["female"]), "person": set([1])}
    right = {"gender": set(["male"]), "person": set([1])}
    with pytest.raises(IncongruentFeature):
        unify_features(left, right)


def test_incongruent_feature_ignores_incongruent_field():
    left = {"gender": set(["female"]), "person": set([1])}
    right = {"gender": set(["male"]), "person": set([1])}
    result = unify_features(left, right, match_required=set(["person"]))
    assert result == {"gender": set(["female", "male"]), "person": set([1])}



from collections import defaultdict
from .production import (
    ProductionBase,
    Production,
    ProductionUnion,
    Terminal
)


class CFG(object):
    """
    a context-free-grammar representation.  this cfg honors
    productions with features, as well.
    """
    def __init__(self, productions):
        self._compiled = defaultdict(list)
        for p in productions:
            self.add(*p)

    def permutations(self, reference):
        production = self._resolve_production(reference)
        return production.permutations(self)

    def permutation_values(self, reference):
        """
        returns just a list of the valid value, instead of
        result objects
        """
        for result in self.permutations(reference):
            yield result.value

    def _resolve_production(self, reference):
        if not isinstance(reference, ProductionBase):
            if reference not in self._compiled:
                raise IndexError("no production {0} found.".format(reference))
            return ProductionUnion(self._compiled[reference])
        return reference

    def add(self, name, production):
        """ add a production to the compiled context """
        self._compiled[name].append(production)



from nlgen.cfg import CFG, Production, ProductionRef, Terminal


def test_full_example():
    my_cfg = CFG([
        ("PRONOUN", Terminal("I")),
        ("VERB", Terminal("have")),
        ("NOUN", Terminal("candy")),
        ("NOUN", Terminal("bonbons")),
        ("SENTENCE", Production([
            ProductionRef("PRONOUN"),
            ProductionRef("VERB"),
            ProductionRef("NOUN")
        ]))
    ])
    result = set(my_cfg.permutation_values("SENTENCE"))
    expect = set([
        ("I", "have", "candy"),
        ("I", "have", "bonbons"),
    ])


def test_feature_mismatch():
    """ a feature mismatch should result """
    my_cfg = CFG([
        ("PRONOUN", Terminal("I", features={"person": set("1")})),
        ("PRONOUN", Terminal("You", features={"person": set("2")})),
        ("VERB", Terminal("have", features={"person": set("2")})),
        ("NOUN", Terminal("candy")),
        ("SENTENCE", Production([
            ProductionRef("PRONOUN"),
            ProductionRef("VERB"),
            ProductionRef("NOUN")
        ]))
    ])
    result = set(my_cfg.permutation_values("SENTENCE"))
    expect = set([("You", "have", "candy")])
    assert result == expect



from nlgen.cfg import CFG, Terminal


def test_multiple_productions_same_rule():
    cfg = CFG([
        ("S", Terminal("foo")),
        ("S", Terminal("bar"))
    ])
    result = cfg.permutation_values("S")
    expect = [("foo",), ("bar",)]
    assert set(result) == set(expect)



"""
it is typically more convenient to use a markup language to define grammars,
instead of using the programmatic interface. This handles the parsing.
"""
from .markup_parser import CFGMarkupParser
from .production import (
    ProductionUnion, Production, ProductionRef, Terminal
)


def parse_production(text):
    """ generate a production from the standard cfg notation of NLGEN. """
    semantics = ProductionSemantics()
    parser = CFGMarkupParser(semantics=semantics)
    ast = parser.parse(text, "production")
    return ast


class ProductionSemantics(object):

    def terminal(self, ast):
        return Terminal(ast.value)

    def reference(self, ast):
        return ProductionRef(ast.key)

    def production(self, ast):
        return Production(ast.production_list)



#!/usr/bin/env python

try:
    from setuptools import setup, find_packages
except:
    from distutils.core import setup

setup(name='nlgen',
      version='0.0.3',
      description='natural language generator',
      long_description="the language game.",
      author='Yusuke Tsutsumi',
      author_email='yusuke@tsutsumi.io',
      packages=find_packages(),
      install_requires=[
          "grako"
      ],
      classifiers=[
          'Programming Language :: Python :: 3',
      ],
      entry_points={
          'console_scripts': []
      },
      tests_require=[]
)



"""
all exceptions belong here.
"""


class NLGenException(Exception):
    """ All exceptions should inherit from this exception. """
    pass


class IncongruentFeature(Exception):
    """
    raised when a feature set can be merged
    with another feature set.
    """



from nlgen.cfg import CFG, Terminal


def test_simple_terminal():
    cfg = CFG([
        ("S", Terminal("foo"))
    ])
    assert list(cfg.permutation_values("S")) == [("foo",)]


def test_coerce_features():
    """ we should be able to coerce common feature descriptions. """
    # string as value
    assert Terminal("foo", features={"num": "1"}).features == {"num": {"1"}}
    # list as value
    assert Terminal("foo", features={"num": ["1", "2"]}).features == {"num": {"1", "2"}}
    # coerce all else as strings.
    assert Terminal("foo", features={"num": 1}).features == {"num": {"1"}}



from .result import Result
from ..exception import (
    IncongruentFeature,
)
from .feature import unify_features
from ..compat import string_type


class ProductionBase(object):
    pass


class ProductionUnion(ProductionBase):
    """
    the union of several productions. It will
    iterate through all matches through all production lists.
    """

    def __init__(self, individual_production_list):
        self._individual_production_list = individual_production_list

    def permutations(self, cfg):
        for production in self._individual_production_list:
            for permutation in production.permutations(cfg):
                yield permutation


class Production(ProductionBase):
    """
    a list of productions, such as
    PRONOUN " " VERB " " NOUN
    """

    def __init__(self, production_list):
        self._production_list = production_list

    def permutations(self, cfg):
        return self._permutations_from_list(
            self._production_list, cfg
        )

    def _permutations_from_list(self, production_list, cfg):
        if len(production_list) == 0:
            yield Result((), {})
        else:
            lhs = production_list[0]
            rhs = production_list[1:]
            for lhs_value in lhs.permutations(cfg):
                for rhs_value in self._permutations_from_list(rhs, cfg):
                    try:
                        unioned_features = unify_features(lhs_value.features,
                                                          rhs_value.features)
                        yield Result(lhs_value.value + rhs_value.value, unioned_features)
                    except IncongruentFeature:
                        # we don't consider
                        # incongruent features.
                        continue

    def __eq__(self, other):
        if not isinstance(other, Production):
            return False
        return self._production_list == other._production_list


class ProductionRef(ProductionBase):

    def __init__(self, key):
        self._key = key

    def permutations(self, cfg):
        return cfg.permutations(self._key)

    def __eq__(self, other):
        return (
            isinstance(other, ProductionRef) and
            self._key == other._key
        )


class Terminal(ProductionBase):

    def __init__(self, value, features=None):
        self._value = value
        self._features = self._coerce_features(features or {})

    @property
    def features(self):
        return self._features

    def permutations(self, cfg):
        yield Result((self._value,), self._features)

    def __eq__(self, other):
        return (
            isinstance(other, Terminal) and
            self._value == other._value
        )

    def __repr__(self):
        return "<Terminal: {0}>".format(self._value.encode("utf-8"))

    def __hash__(self):
        return hash(self._value)

    @staticmethod
    def _coerce_features(feature_dict):
        """
        the features dictionary can accept a variety
        of types, so this method provides proper
        enforcement to the correct data structure.

        the output data structure is a dictionary
        of <str, set> mappings.
        """
        for k, v in feature_dict.items():
            if isinstance(v, set):
                # the desired type.
                continue
            if isinstance(v, string_type):
                v = set([v])
            elif hasattr(v, "__iter__"):
                v = set(v)
            else:
                v = set([str(v)])
            feature_dict[k] = v
        return feature_dict



from nlgen.cfg import CFG, Terminal, ProductionUnion


def test_simple_production_union():
    cfg = CFG([
        ("S", ProductionUnion([
            Terminal("foo"),
            Terminal("bar")
        ])),
    ])
    expect = [("foo",), ("bar",)]
    result = list(cfg.permutation_values("S"))
    assert expect == result
